import os
import string
import random
import unittest
from HW3 import *

class Test_Ngrams(unittest.TestCase):

    def test_n_grams(self):
        self.assertEqual(n_grams(1, "سلام"), [("~", "س"), ("س", "ل"), ("ل", "ا"), ("ا", "م")])
        self.assertEqual(n_grams(2, "پایان"), [('~~', 'پ'), ('~پ', 'ا'), ('پا', 'ی'), ('ای', 'ا'), ('یا', 'ن')])

    def test_examples(self):

        model = N_Gram_Model(1, 0)
        model.update("abab")
        self.assertEqual(model.get_vocabulary(), {"a", "b"})
        model.update("abcd")
        self.assertEqual(model.get_vocabulary(), {"a", "b", "c", "d"})
        self.assertEqual(model.probability("a", "b"), 1.0)
        self.assertEqual(model.probability("~", "c"), 0.0)
        self.assertEqual(model.probability("b", "c"), 0.5)

        model_2 = N_Gram_Model(1, 0)
        model_2.update("سلام")
        self.assertEqual(model_2.get_vocabulary(), {"س", "ل", "ا", "م"})
        model_2.update("پایان")
        self.assertEqual(model_2.get_vocabulary(), {"س", "ل", "ا", "م", "پ", "ی", "ن"})
        self.assertEqual(model_2.probability("س", "ل"), 1.0)
        self.assertEqual(model_2.probability("~", "پ"), 0.5)
        self.assertEqual(model_2.probability("ا", "ن"), 0.3333333333333333)

    def test_zero_gram(self):
        model = N_Gram_Model(0, 0)
        model.update("سلام")
        model.update("پایان")
        random.seed(1) # for testing purposes
        self.assertEqual([model.random_character('') for i in range(25)], ['ا', 'پ', 'ن', 'ا', 'ل', 'ل', 'م', 'پ', 'ا', 'ا', 'پ', 'س', 'ن', 'ا', 'ل', 'ن', 'ا', 'ی', 'ی', 'ا', 'ا', 'ل', 'ی', 'س', 'ا'])

    def test_random_text(self):
        model = N_Gram_Model(1, 0)
        model.update("abab")
        model.update("abcd")
        random.seed(1)
        self.assertEqual(model.generate_random_text(25), "abcdbabcdabababcdddabcdba")

    def test_random_text_FA(self):
        model = N_Gram_Model(1, 0)
        model.update("سلام")
        model.update("پایان")
        random.seed(1)
        self.assertEqual(model.generate_random_text(25), "سلامممنپامپایانپایامانیام")

    def test_model(self):
        random.seed(42)
        n = 9
        model = create_ngram_model(N_Gram_Model, 'poems.txt', n)
        output_file = 'poems_output_{}.txt'.format(n)
        if os.path.exists(output_file):
            os.remove(output_file)
        with open(output_file, 'w', encoding='utf-8') as f:
            f.write(model.generate_random_text(250))
        self.assertTrue(os.path.exists(output_file))
        with open(output_file, 'r', encoding='utf-8') as f:
            generated_text = f.read()
        random.seed(42)
        expected_text = model.generate_random_text(250)
        self.assertEqual(generated_text, expected_text)
    
    def test_empty_input(self):
        model = N_Gram_Model(1, 0)
        model.update('')
        self.assertEqual(model.perplexity(''), float('inf'))
        
    def test_single_char_input(self):
        model = N_Gram_Model(1, 0)
        model.update('a')
        self.assertEqual(model.perplexity('a'), 1)
    
    def test_unseen_ngram(self):
        model = N_Gram_Model(2, 0)
        model.update('abc')
        self.assertEqual(model.perplexity('def'), float('inf'))
    
    def test_different_order_ngram(self):
        model_1 = N_Gram_Model(1, 0)
        model_1.update('abc')
        
        model_2 = N_Gram_Model(2, 0)
        model_2.update('abc')

        self.assertEqual(model_2.perplexity('def'), model_1.perplexity('def'))
        
    def test_large_input(self):
        model = N_Gram_Model(3, 0)
        input_str = ''.join(random.choice(string.ascii_lowercase) for _ in range(10000))
        model.update(input_str)
        self.assertLess(model.perplexity(input_str), float('inf'))

    def test_perplexity(self):
        model = N_Gram_Model(1, 0)
        model.update('abab')
        model.update('abcd')
        self.assertEqual(model.perplexity('abca'), float('inf'))
        self.assertEqual(model.perplexity('abcda'), 1.515716566510398)

    def test_prob(self):
        model_1 = N_Gram_Model(1, 1)
        model_1.update('abab')
        model_1.update('abcd')
        self.assertAlmostEqual(model_1.probability('a', 'a'), 0.14285714285714285)
        self.assertAlmostEqual(model_1.probability('a', 'b'), 0.5714285714285714)
        self.assertAlmostEqual(model_1.probability('c', 'd'), 0.4)
        self.assertAlmostEqual(model_1.probability('d', 'a'), 0.25)
        
        model_2 = N_Gram_Model(2, 1)
        model_2.update('abab')
        model_2.update('abcd')
        self.assertAlmostEqual(model_2.probability('ab', 'a'), 0.3333333333333333)
        self.assertAlmostEqual(model_2.probability('ab', 'b'), 0.16666666666666666)
        self.assertAlmostEqual(model_2.probability('bc', 'd'), 0.4)
        self.assertAlmostEqual(model_2.probability('cd', 'a'), 0.25)
        
        model_3 = N_Gram_Model(3, 1)
        model_3.update('abab')
        model_3.update('abcd')
        self.assertAlmostEqual(model_3.probability('aba', 'b'), 0.4)
        self.assertAlmostEqual(model_3.probability('bcd', 'a'), 0.25)
        self.assertAlmostEqual(model_3.probability('abc', 'd'), 0.4)
        self.assertAlmostEqual(model_3.probability('bca', 'b'), 0.25)
        
        model_4 = N_Gram_Model(2, 0)
        model_4.update('aaabbbccc')
        self.assertAlmostEqual(model_4.probability('aa', 'a'), 0.5)
        self.assertAlmostEqual(model_4.probability('ab', 'b'), 1.0)
        
        model_5 = N_Gram_Model(1, 0)
        model_5.update('aaabbbccc')
        self.assertAlmostEqual(model_5.probability('a', 'a'), 0.6666666666666666)
        self.assertAlmostEqual(model_5.probability('c', 'c'), 1.0)

    # Using uniform distribution for weights
    def test_prob_interpolation(self):
        model_1 = N_Gram_Interpolation(1, 0)
        model_1.update('abab')
        self.assertAlmostEqual(model_1.probability('a', 'a'), 0.25)
        self.assertAlmostEqual(model_1.probability('a', 'b'), 0.75)
        
        model_2 = N_Gram_Interpolation(2, 1)
        model_2.update('abab')
        model_2.update('abcd')
        self.assertAlmostEqual(model_2.probability('~a', 'b'), 0.4682539682539682)
        self.assertAlmostEqual(model_2.probability('ba', 'b'), 0.4349206349206349)
        self.assertAlmostEqual(model_2.probability('~c', 'd'), 0.27222222222222225)
        self.assertAlmostEqual(model_2.probability('bc', 'd'), 0.3222222222222222)
        
        model_3 = N_Gram_Interpolation(3, 1)
        model_3.update('abab')
        model_3.update('abcd')
        self.assertAlmostEqual(model_3.probability('aba', 'b'), 0.4261904761904761)
        self.assertAlmostEqual(model_3.probability('bcd', 'a'), 0.2708333333333333)
        self.assertAlmostEqual(model_3.probability('abc', 'd'), 0.3416666666666667)
        self.assertAlmostEqual(model_3.probability('bca', 'b'), 0.35119047619047616)
        
        model_4 = N_Gram_Interpolation(2, 0)
        model_4.update('aaabbbccc')
        self.assertAlmostEqual(model_4.probability('aa', 'a'), 0.5)
        self.assertAlmostEqual(model_4.probability('ab', 'b'), 0.6666666666666666)
        
        model_5 = N_Gram_Interpolation(1, 0)
        model_5.update('aaabbbccc')
        self.assertAlmostEqual(model_5.probability('a', 'a'), 0.5)
        self.assertAlmostEqual(model_5.probability('c', 'c'), 0.6666666666666666)


if __name__ == "__main__":
    unittest.main()